/*
 * Copyright (C) 2013 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.android.builder.compiling;

import com.android.annotations.NonNull;

import java.io.File;

/**
 * A Class that processes a dependency file after a compilation.
 *
 * During compilation of aidl and renderscript, it is possible to provide an instance of
 * DependencyFileProcessor to process the dependency files generated by the compilers.
 *
 * It can be useful to store the dependency in a better format than a per-file dependency file.
 *
 * The instance will be called for each dependency file that is created during compilation, and
 * if the file can be processed will notify the compiler that the original dependency file is not
 * needed anymore.
 *
 * @see com.android.builder.AndroidBuilder#compileAllAidlFiles(java.util.List, java.io.File, java.util.List, DependencyFileProcessor)
 * @see com.android.builder.AndroidBuilder#compileAidlFile(java.io.File, java.io.File, java.util.List, DependencyFileProcessor)
 * @see com.android.builder.AndroidBuilder#compileAllRenderscriptFiles(java.util.List, java.util.List, java.io.File, java.io.File, int, boolean, int)
 */
public interface DependencyFileProcessor {

    /**
     * Processes the dependency file.
     * @param dependencyFile the dependency file.
     * @return true if the dependency file can be deleted by the caller.
     */
    boolean processFile(@NonNull File dependencyFile);
}
